from nacl.signing import SigningKey, VerifyKey

from leapchain.accounts.manage import create_account


def test_create_account():
    signing_key, account_number = create_account()

    assert isinstance(signing_key, SigningKey)
    assert isinstance(account_number, VerifyKey)



import factory


def build_json(factory_class, **kwargs):
    """Build json representation for object using factory."""
    return factory.build(
        dict,
        FACTORY_CLASS=factory_class,
        **kwargs,
    )



def validate_transaction_exists(*, amount, fee, error, recipient, txs):
    """Check for the existence of a Tx"""
    tx = next(
        (
            tx for tx in txs if
            tx.get('amount') >= amount
            and tx.get('fee') == fee
            and tx.get('recipient') == recipient
        ),
        None
    )

    if not tx:
        raise error({
            'error_message': 'Tx not found',
            'expected_amount': amount,
            'expected_fee': fee,
            'expected_recipient': recipient
        })



from rest_framework import serializers

from leapchain.constants.network import BLOCK_IDENTIFIER_LENGTH
from leapchain.serializers.account_balance import AccountBalanceSerializer
from leapchain.serializers.network_block import NetworkBlockSerializer


class ConfirmationBlockMessageSerializer(serializers.Serializer):
    block = NetworkBlockSerializer()
    block_identifier = serializers.CharField(max_length=BLOCK_IDENTIFIER_LENGTH)
    updated_balances = AccountBalanceSerializer(many=True)

    def create(self, validated_data):
        pass

    def update(self, instance, validated_data):
        pass



from django.core.exceptions import ValidationError
from django.core.management import CommandParser

from leapchain.argparser.validators import decimal_validator
from leapchain.base_classes.initialize_node import InitializeNode
from leapchain.constants.network import PRIMARY_VALIDATOR
from leapchain.utils.format import format_address
from leapchain.utils.network import fetch
from leapchain.utils.validators import validate_is_real_number

"""
The FetchPrimaryValidatorConfig class contains logic to fetch and validate configuration data from a primary validator.
"""


class FetchPrimaryValidatorConfig(InitializeNode):

    def __init__(self, *args, **kwargs):
        """Initialize FetchPrimaryValidatorConfig class"""
        super().__init__(*args, **kwargs)

        self.required_input = {
            'ip_address': None,
            'port': None,
            'protocol': None,
            'trust': None
        }

    def add_arguments(self, parser: CommandParser):
        """Additional custom arguments"""
        super().add_arguments(parser)
        parser.add_argument('--trust', type=decimal_validator(min_val=0, max_val=100))

    def get_primary_validator_address(self):
        """Return formatted address of primary validator"""
        return format_address(
            ip_address=self.required_input['ip_address'],
            port=self.required_input['port'],
            protocol=self.required_input['protocol']
        )

    def get_trust(self, value=None):
        """Get trust from user"""
        valid = False

        while not valid:
            if self.unattended:
                trust = value
            else:
                trust = input('Enter trust (required): ')

            if not trust:
                self._error('trust required')
                continue

            is_valid_decimal, trust = self.validate_and_convert_to_decimal(trust)

            if not is_valid_decimal:
                continue

            try:
                validate_is_real_number(trust)
            except ValidationError:
                self._error('Value must be a real number')
                continue

            if trust < 0:
                self._error('Value can not be less than 0')
                continue

            if trust > 100:
                self._error('Value can not be greater than 100')
                continue

            self.required_input['trust'] = trust
            valid = True

    def fetch_validator_config(self):
        """Return config"""
        address = self.get_primary_validator_address()
        url = f'{address}/config'
        results = fetch(url=url, headers={})
        return results

    def handle(self, *args, **options):
        """Run script"""
        connected = False

        while not connected:
            self.required_input = {
                'ip_address': None,
                'port': None,
                'protocol': None
            }

            self.get_ip_address(value=options.get('ip_address'))
            self.get_protocol(value=options.get('protocol'))
            self.get_port(value=options.get('port'))

            try:
                config = self.fetch_validator_config()

                if not self.is_config_valid(config):
                    continue

                self.get_trust(value=options.get('trust'))
                self.handle_primary_validator_config(config)
            except Exception as e:
                self._error('Unable to connect')
                self._error(e)
                continue

            connected = True

        self.stdout.write(self.style.SUCCESS('Success'))

    def handle_primary_validator_config(self, primary_validator_config):
        """Handle primary validator configuration data"""
        raise NotImplementedError(
            'subclasses of FetchPrimaryValidatorConfig must provide a handle_primary_validator_config() method'
        )

    def is_config_valid(self, config):
        """Validate config response data from the validator"""
        if config.get('node_type') != PRIMARY_VALIDATOR:
            self._error(f'node_type is not {PRIMARY_VALIDATOR}')
            return False

        return True



def assert_objects_vs_dicts(objects, dicts, sort=True, key='id'):
    """Compare list of django objects with serialized objects using their id's"""
    objects_ids = [str(getattr(obj, key)) for obj in objects]
    dicts_ids = [str(obj.get(key)) for obj in dicts]

    if sort:
        assert sorted(objects_ids) == sorted(dicts_ids)
    else:
        assert objects_ids == dicts_ids



from rest_framework import serializers

from leapchain.blocks.signatures import verify_signature
from leapchain.constants.network import BANK, PRIMARY_VALIDATOR, SIGNATURE_LENGTH, VERIFY_KEY_LENGTH
from leapchain.serializers.message import MessageSerializer
from leapchain.utils.serializers import validate_keys
from leapchain.utils.tools import sort_and_encode


class NetworkBlockSerializer(serializers.Serializer):
    account_number = serializers.CharField(max_length=VERIFY_KEY_LENGTH, min_length=VERIFY_KEY_LENGTH)
    message = MessageSerializer()
    signature = serializers.CharField(max_length=SIGNATURE_LENGTH, min_length=SIGNATURE_LENGTH)

    def create(self, validated_data):
        pass

    def update(self, instance, validated_data):
        pass

    def validate(self, data):
        """Validate signature, unique Tx recipients, unique Tx fees and account_number not included as a Tx recipient"""
        account_number = data['account_number']
        message = data['message']
        txs = message['txs']
        signature = data['signature']

        verify_signature(
            message=sort_and_encode(message),
            signature=signature,
            verify_key=account_number
        )

        recipient_list = [tx['recipient'] for tx in txs]
        recipient_set = set(recipient_list)

        if len(recipient_list) != len(recipient_set):
            raise serializers.ValidationError('Tx recipients must be unique')

        if account_number in recipient_set:
            raise serializers.ValidationError('Block account_number not allowed as Tx recipient')

        bank_fee_exists = False
        primary_validator_fee_exists = False

        for tx in txs:
            fee = tx.get('fee', None)

            if fee is None:
                continue

            if fee == BANK:

                if bank_fee_exists:
                    raise serializers.ValidationError('Multiple bank fees not allowed')
                else:
                    bank_fee_exists = True

            if fee == PRIMARY_VALIDATOR:

                if primary_validator_fee_exists:
                    raise serializers.ValidationError('Multiple primary validator fees not allowed')
                else:
                    primary_validator_fee_exists = True

        validate_keys(self, data)

        return data



from rest_framework import serializers

from leapchain.constants.network import BALANCE_LOCK_LENGTH
from leapchain.serializers.network_transaction import NetworkTransactionSerializer
from leapchain.utils.serializers import validate_keys


class MessageSerializer(serializers.Serializer):
    balance_key = serializers.CharField(max_length=BALANCE_LOCK_LENGTH)
    txs = NetworkTransactionSerializer(many=True)

    def create(self, validated_data):
        pass

    def update(self, instance, validated_data):
        pass

    def validate(self, data):
        """Validate Txs exist"""
        if not data['txs']:
            raise serializers.ValidationError('Invalid Txs')

        validate_keys(self, data)

        return data



from uuid import uuid4

from django.db import models

from leapchain.models.created_modified import CreatedModified


class ConfirmationService(CreatedModified):
    id = models.UUIDField(default=uuid4, editable=False, primary_key=True)  # noqa: A003
    end = models.DateTimeField()
    start = models.DateTimeField()

    class Meta:
        abstract = True



from hashlib import sha3_256 as sha3

from leapchain.utils.tools import sort_and_encode


def get_message_hash(*, message):
    """Return has of given message"""
    return sha3(sort_and_encode(message)).digest().hex()



from factory import Faker
from factory.django import DjangoModelFactory

from ..models.created_modified import CreatedModified


class CreatedModifiedFactory(DjangoModelFactory):
    created_date = Faker('date')
    modified_date = Faker('date')

    class Meta:
        model = CreatedModified
        abstract = True



from uuid import uuid4

from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models

from leapchain.constants.network import MAX_POINT_VALUE, MIN_POINT_VALUE, PROTOCOL_CHOICES, VERIFY_KEY_LENGTH


class NetworkNode(models.Model):
    id = models.UUIDField(default=uuid4, editable=False, primary_key=True)  # noqa: A003
    account_number = models.CharField(max_length=VERIFY_KEY_LENGTH)
    ip_address = models.GenericIPAddressField()
    node_identifier = models.CharField(max_length=VERIFY_KEY_LENGTH, unique=True)
    port = models.PositiveIntegerField(blank=True, null=True, validators=[MaxValueValidator(65535)])
    protocol = models.CharField(choices=PROTOCOL_CHOICES, max_length=5)
    version = models.CharField(max_length=32)

    # Fees
    default_transaction_fee = models.PositiveBigIntegerField(
        default=MIN_POINT_VALUE,
        validators=[
            MaxValueValidator(MAX_POINT_VALUE),
            MinValueValidator(MIN_POINT_VALUE),
        ]
    )

    class Meta:
        abstract = True
        indexes = [
            models.Index(fields=['ip_address', 'port', 'protocol']),
        ]
        constraints = [
            models.UniqueConstraint(
                fields=['ip_address', 'port', 'protocol'],
                name='%(app_label)s_%(class)s_unique_ip_port_proto')
        ]



import math

from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _


def validate_is_real_number(value):
    """Validate value is real number, does not allow Infinity, -Infinity, or NaN"""
    if not value:
        return

    if math.isnan(value):
        raise ValidationError(
            _('NaN is not allowed'),
            params={'value': value},
        )

    if math.isinf(value):
        raise ValidationError(
            _('Infinity is not allowed'),
            params={'value': value},
        )



from nacl.encoding import HexEncoder
from nacl.signing import VerifyKey


def generate_signature(*, message, signing_key):
    """Sign message using signing key and return signature"""
    return signing_key.sign(message).signature.hex()


def verify_signature(*, message, signature, verify_key):
    """Verify block signature"""
    verify_key = VerifyKey(verify_key.encode('utf-8'), encoder=HexEncoder)
    signature = bytes.fromhex(signature)
    verify_key.verify(message, signature)



HTTP = 'http'
HTTPS = 'https'

PROTOCOL_CHOICES = [
    (HTTP, HTTP),
    (HTTPS, HTTPS),
]

PROTOCOL_LIST = [HTTP, HTTPS]

BANK = 'BANK'
CONFIRMATION_VALIDATOR = 'CONFIRMATION_VALIDATOR'
PRIMARY_VALIDATOR = 'PRIMARY_VALIDATOR'

ACCOUNT_FILE_HASH_LENGTH = 64
BALANCE_LOCK_LENGTH = 64
BLOCK_IDENTIFIER_LENGTH = 64
HEAD_HASH_LENGTH = 64
SIGNATURE_LENGTH = 128
VERIFY_KEY_LENGTH = 64

MEMO_MAX_LENGTH = 256

MIN_POINT_VALUE = 1
MAX_POINT_VALUE = 281474976710656

ACCEPTED_FEE_CHOICES = [
    (BANK, BANK),
    (PRIMARY_VALIDATOR, PRIMARY_VALIDATOR),
]

ACCEPTED_FEE_LIST = [BANK, PRIMARY_VALIDATOR]



def all_field_names(model):
    """Return list of all field names for model"""
    return [f.name for f in model._meta.get_fields()]


def common_field_names(a, b):
    """Return field names that are common between different models"""
    field_names_a = standard_field_names(a)
    field_names_b = standard_field_names(b)
    return field_names_a.intersection(field_names_b)


def standard_field_names(model):
    """Return set of standard (excluding date, ID, and relational) field names for model"""
    excluded = ['id', 'created_date', 'modified_date']
    return {f.name for f in model._meta.get_fields() if f.name not in excluded and not f.is_relation}



import decimal

from django.core.exceptions import ValidationError
from django.core.management.base import BaseCommand, CommandParser
from django.core.validators import validate_ipv46_address

from leapchain.argparser.validators import int_validator, ipv46_validator, str_length_validator
from leapchain.constants.network import PROTOCOL_LIST, VERIFY_KEY_LENGTH

"""
The InitializeNode class contains the core logic used when initializing network nodes via the command line.

Includes methods to:
- validate config settings are properly formatted
- standardize error handling
"""


class InitializeNode(BaseCommand):
    unattended = False

    def __init__(self, *args, **kwargs):
        """Inits InitializeNode class"""
        super().__init__(*args, **kwargs)
        self.required_input = {}

    def add_arguments(self, parser: CommandParser):
        """Additional custom arguments"""
        super().add_arguments(parser)
        parser.add_argument('--node_identifier', type=str_length_validator(length=VERIFY_KEY_LENGTH))
        parser.add_argument('--account_number', type=str_length_validator(length=VERIFY_KEY_LENGTH))
        parser.add_argument('--default_transaction_fee', type=int)
        parser.add_argument('--ip_address', type=ipv46_validator())
        parser.add_argument('--port', type=int_validator(min_val=0, max_val=65535))
        parser.add_argument('--protocol', choices=PROTOCOL_LIST)
        parser.add_argument('--version_number', type=str_length_validator(max_len=32))

        parser.add_argument('--unattended', action='store_true')

    def _error(self, message):
        """Display error message string in console"""
        self.stdout.write(self.style.ERROR(message))

    def get_fee(self, *, attribute_name, human_readable_name, value=None):
        """
        Validate fee

        - default_transaction_fee
        """
        valid = False

        while not valid:
            if self.unattended:
                fee = value
            else:
                fee = input(f'Enter {human_readable_name} (required): ')

            if not fee:
                self._error(f'{attribute_name} required')
                continue

            is_valid_integer, fee = self.validate_and_convert_to_integer(fee)

            if not is_valid_integer:
                continue

            self.required_input[attribute_name] = fee
            valid = True

    def get_ip_address(self, value=None):
        """Get IP address from user"""
        valid = False

        while not valid:
            if self.unattended:
                ip_address = value
            else:
                ip_address = input('Enter public IP address (required): ')

            if not ip_address:
                self._error('ip_address required')
                continue

            try:
                validate_ipv46_address(ip_address)
            except ValidationError:
                self._error('Enter a valid IPv4 or IPv6 address')
                continue

            self.required_input['ip_address'] = ip_address
            valid = True

    def get_port(self, value=None):
        """Get port from user"""
        valid = False

        while not valid:
            if self.unattended:
                port = value
            else:
                port = input('Enter port: ')

            if not port:
                break

            try:
                port = int(port)
            except ValueError:
                self._error(f'{port} is not a valid integer')
                continue

            if port < 0:
                self._error('port can not be less than 0')
                continue

            if port > 65535:
                self._error('port can not be greater than 65535')
                continue

            self.required_input['port'] = port
            valid = True

    def get_protocol(self, value=None):
        """Get protocol from user"""
        valid = False

        while not valid:
            if self.unattended:
                protocol = value
            else:
                protocol = input('Enter protocol (required): ')

            if not protocol:
                self._error('protocol required')
                continue

            if protocol not in PROTOCOL_LIST:
                self._error(f'protocol must be one of {PROTOCOL_LIST}')
                continue

            self.required_input['protocol'] = protocol
            valid = True

    def get_verify_key(self, *, attribute_name, human_readable_name, value=None):
        """
        Validate verify key

        - account_number
        - node_identifier
        """
        valid = False

        while not valid:
            if self.unattended:
                verify_key = value
            else:
                verify_key = input(f'Enter {human_readable_name} (required): ')

            if not verify_key:
                self._error(f'{attribute_name} required')
                continue

            if len(verify_key) != VERIFY_KEY_LENGTH:
                self._error(f'{attribute_name} must be {VERIFY_KEY_LENGTH} characters long')
                continue

            self.required_input[attribute_name] = verify_key
            valid = True

    def get_version_number(self, value=None):
        """Get version from user"""
        max_length = 32
        valid = False

        while not valid:
            if self.unattended:
                version = value
            else:
                version = input('Enter version (required): ')

            if not version:
                self._error('version required')
                continue

            if len(version) > max_length:
                self._error(f'version must be less than or equal to {max_length} characters long')
                continue

            self.required_input['version'] = version
            valid = True

    def execute(self, *args, **options):
        self.unattended = options.get('unattended', False)
        return super().execute(*args, **options)

    def handle(self, *args, **options):
        """The actual logic of the command. Subclasses must implement this method"""
        raise NotImplementedError('subclasses of InitializeNode must provide a handle() method')

    def validate_and_convert_to_decimal(self, value):
        """
        Validate that given value can be converted to Decimal value

        Returns: is_valid (bool), decimal_value (Decimal)
        Must return is_valid flag along with decimal_value to allow for proper validation of valid falsy value (0.0)
        """
        try:
            value = decimal.Decimal(value)
        except decimal.InvalidOperation:
            self._error(f'Can not convert {value} to a decimal')
            return False, None

        return True, value

    def validate_and_convert_to_integer(self, value):
        """
        Validate that given value can be converted to integer value

        Returns: is_valid (bool), integer_value (int)
        """
        try:
            value = int(value)
        except ValueError:
            self._error(f'Can not convert {value} to an integer')
            return False, None

        return True, value



from os import path

from nacl.encoding import HexEncoder
from nacl.signing import SigningKey

from leapchain.accounts.manage import create_account


def create_account_and_save_signing_key_file(file):
    """
    Create a new account and save signing key to file

    Return signing_key, account_number
    """
    signing_key, account_number = create_account()
    write_signing_key_file(signing_key, file)
    return signing_key, account_number


def read_signing_key_file(file):
    """Read signing key from file"""
    with open(file, 'rb') as f:
        return SigningKey(f.read(), encoder=HexEncoder)


def write_signing_key_file(signing_key, file):
    """Save signing key to file"""
    if not isinstance(signing_key, SigningKey):
        raise RuntimeError('signing_key must be of type nacl.signing.SigningKey')

    if path.exists(file):
        raise RuntimeError(f'{file} already exists')

    with open(file, 'wb') as f:
        f.write(signing_key.encode(encoder=HexEncoder))



from django.db import models

from leapchain.constants.network import ACCOUNT_FILE_HASH_LENGTH, BLOCK_IDENTIFIER_LENGTH
from leapchain.models.network_node import NetworkNode


class NetworkValidator(NetworkNode):
    root_account_file = models.URLField(max_length=1024)
    root_account_file_hash = models.CharField(max_length=ACCOUNT_FILE_HASH_LENGTH)
    seed_block_identifier = models.CharField(blank=True, max_length=BLOCK_IDENTIFIER_LENGTH)

    # Confirmation rate (used by confirmation validators only)
    daily_confirmation_rate = models.PositiveBigIntegerField(blank=True, null=True)

    class Meta(NetworkNode.Meta):
        abstract = True



# -*- coding: utf-8 -*-
import re

from setuptools import find_packages, setup

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

with open('requirements.txt', encoding='utf-8') as f:
    requirements = f.read().splitlines()

with open('src/leapchain/__init__.py', encoding='utf8') as f:
    version = re.search(r'__version__ = \'(.*?)\'', f.read()).group(1)

setup(
    install_requires=requirements,
    long_description=long_description,
    long_description_content_type='text/markdown',
    name='leapchain',
    packages=find_packages(
        exclude=['tests', 'tests.*']
    ),
    version=version,
)



from factory import Faker
from factory.django import DjangoModelFactory
from factory.fuzzy import FuzzyChoice

from ..constants.network import MAX_POINT_VALUE, MIN_POINT_VALUE, PROTOCOL_CHOICES, VERIFY_KEY_LENGTH
from ..models.network_node import NetworkNode


class NetworkNodeFactory(DjangoModelFactory):
    account_number = Faker('text', max_nb_chars=VERIFY_KEY_LENGTH)
    default_transaction_fee = Faker('pyint', max_value=MAX_POINT_VALUE, min_value=MIN_POINT_VALUE)
    ip_address = Faker('ipv4')
    node_identifier = Faker('text', max_nb_chars=VERIFY_KEY_LENGTH)
    port = Faker('random_int', min=0, max=65535)
    protocol = FuzzyChoice([p[1] for p in PROTOCOL_CHOICES])
    version = Faker('text', max_nb_chars=32)

    class Meta:
        model = NetworkNode
        abstract = True



from operator import itemgetter

from leapchain.blocks.signatures import generate_signature
from leapchain.utils.tools import sort_and_encode
from leapchain.verify_keys.verify_key import encode_verify_key


def generate_block(*, account_number, balance_lock, signing_key, transactions):
    """Generate block"""
    message = {
        'balance_key': balance_lock,
        'txs': sorted(transactions, key=itemgetter('recipient'))
    }

    signature = generate_signature(
        message=sort_and_encode(message),
        signing_key=signing_key
    )

    block = {
        'account_number': encode_verify_key(verify_key=account_number),
        'message': message,
        'signature': signature
    }

    return block



from rest_framework import serializers

from leapchain.constants.network import MAX_POINT_VALUE, VERIFY_KEY_LENGTH


class AccountBalanceSerializer(serializers.Serializer):
    account_number = serializers.CharField(max_length=VERIFY_KEY_LENGTH)
    balance = serializers.IntegerField(min_value=0, max_value=MAX_POINT_VALUE)

    def create(self, validated_data):
        pass

    def update(self, instance, validated_data):
        pass



from uuid import uuid4

from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models

from leapchain.constants.network import (
    ACCEPTED_FEE_CHOICES,
    MAX_POINT_VALUE,
    MEMO_MAX_LENGTH,
    MIN_POINT_VALUE,
    VERIFY_KEY_LENGTH
)


class NetworkTransaction(models.Model):
    id = models.UUIDField(default=uuid4, editable=False, primary_key=True)  # noqa: A003
    amount = models.PositiveBigIntegerField(
        validators=[
            MaxValueValidator(MAX_POINT_VALUE),
            MinValueValidator(MIN_POINT_VALUE),
        ]
    )
    fee = models.CharField(blank=True, choices=ACCEPTED_FEE_CHOICES, max_length=17)
    memo = models.CharField(blank=True, max_length=MEMO_MAX_LENGTH)
    recipient = models.CharField(max_length=VERIFY_KEY_LENGTH)

    class Meta:
        abstract = True



from django.db import models


class CreatedModified(models.Model):
    created_date = models.DateTimeField(auto_now_add=True, db_index=True, null=True)
    modified_date = models.DateTimeField(auto_now=True, db_index=True)

    class Meta:
        abstract = True
